rm(list=ls())
setwd('../Decomposition')

require(xts)
require(tis)
require(lubridate)
require(pROC)
require(xtable)

load('../Data/Vintage-datasets-Aug-2022.Rda')

data <- rt.vintages
vintage.dates <- as.yearqtr(as.Date(sapply(rt.vintages, function(x) x[['vintage']])))

# These functions produce
source('Weight functions.R')
track.contribution <- function(data.vintages, model.vintages, final=FALSE, quasifinal=FALSE, quasireal=FALSE, min.date) {
  
  # Inputs: data.vintages--either rt.vintages or qrt.vintages
  #         model.vintages--the rt- or qrt- model vintages
  #         final -- a flag if you want to decompose the final vintage output gap estimate
  #         min.date -- If you want to subset the data you are decomposing 
  
  # Output: A matrix with columns corresponding to data inputs and the estimated cycle, c
  #         Rows are the quarter being decomposed!
  #         Notice that the sum of the data inputs across rows is that row's cycle!!
  
  V <- length(model.vintages)
  data.names <- model.vintages[[V]][['prior']][['obs.names']]
  
  taus <- which(vintage.dates > min.date)
  # Eliminate 1996Q1 vintage because BEA did not publish GDP data (government shutdown)
  taus = taus[-122]
  
  N_draws = length(model.vintages[[1]]$res$phi[1,])
  temp_draws = array(NA, dim= c(length(taus), length(data.names)+1,N_draws))
  temp_contr_index <- xts(array(NA, c(length(taus), length(data.names)+1)), 
                          order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
  colnames(temp_contr_index) <- c(data.names, 'c')
  
  i <- min(which(sapply(model.vintages, function(x) length(x))!=0)) #Index for storage of functions
  
  if (final==TRUE) {
    cat("\n ")
    cat("Final true")
    cat("\n ")
    
    # For final, decompose the sequence of output gap estimates from the final vintage
    data$y <- data.vintages[[V]][['y']][,data.names]
    prior <- model.vintages[[V]]$prior
    if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
    c <- which(prior$state.names=='c') # Index for cycle
    for (i_draws in 1:N_draws) {
      
      param = list()
      param$phi = model.vintages[[V]]$res$phi[,i_draws]
      param$omega = model.vintages[[V]]$res$omega[,i_draws]
      if (length(model.vintages[[V]]$res$alpha[,i_draws]) != 0) param$alpha = model.vintages[[V]]$res$alpha[,i_draws]
      if (length(model.vintages[[V]]$res$beta[,i_draws]) != 0) param$beta = model.vintages[[V]]$res$beta[,i_draws]
      
      model <- list()
      model$F <- model.vintages[[i]]$build.F(prior, param)
      model$G <- model.vintages[[i]]$build.G(prior, param)
      model$H <- model.vintages[[i]]$build.H(prior)
      model$omega <- diag(param$omega)
      
      K <- KFKS(data$y, prior, model)
      
      for (t in 1:length(taus)) {
        i2 <- which(index(data$y) == index(temp_contr_index[t]))
        w <- w.stT(data$y, prior, model, K, i2)
        W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
        temp_draws[t, ,i_draws] <- cbind(W, w$stT)[c,]
      } 
    }
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')
    
  } else if (quasifinal==TRUE) {  
    cat("\n ")
    cat("quasifinal true \n")
    cat("\n ")
    
    for (i_draws in 1:N_draws) {
      for (t in 1:length(taus)) {
        param = list()
        param$phi = model.vintages[[V]]$res$phi[,i_draws]
        param$omega = model.vintages[[V]]$res$omega[,i_draws]
        if (length(model.vintages[[V]]$res$alpha[,i_draws]) != 0) param$alpha = model.vintages[[V]]$res$alpha[,i_draws]
        if (length(model.vintages[[V]]$res$beta[,i_draws]) != 0) param$beta = model.vintages[[V]]$res$beta[,i_draws]
        data$y <- data.vintages[[taus[t]]][['y']][,data.names]
        
        prior <- model.vintages[[taus[t]]]$prior
        if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
        c <- which(prior$state.names=='c') # Index for cycle
        
        model <- list()
        model$F <- model.vintages[[i]]$build.F(prior, param)
        model$G <- model.vintages[[i]]$build.G(prior, param)
        model$H <- model.vintages[[i]]$build.H(prior)
        model$omega <- diag(param$omega)
        
        K <- KFKS(data$y, prior, model)
        w <- w.stt(data$y, prior, model, K, dim(K$KS$stT)[1])
        
        W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
        temp_draws[t, ,i_draws] <- cbind(W, w$stt)[c,]
      }
    }
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')

  } else if (quasireal==TRUE) {  
    cat("\n ")
    cat("quasireal true \n")
    cat("\n ")
    
    for (i_draws in 1:N_draws) {
      for (t in 1:length(taus)) {
        S = model.vintages[[taus[t]]]$res$S[,4,i_draws]
        param = list()
        param$phi = model.vintages[[taus[t]]]$res$phi[,i_draws]
        param$omega = model.vintages[[taus[t]]]$res$omega[,i_draws]
        if (length(model.vintages[[taus[t]]]$res$alpha[,i_draws]) != 0) param$alpha = model.vintages[[taus[t]]]$res$alpha[,i_draws]
        if (length(model.vintages[[taus[t]]]$res$beta[,i_draws]) != 0) param$beta = model.vintages[[taus[t]]]$res$beta[,i_draws]
        if (length(S)!=0) {
          
          data$y <- data.vintages[[taus[t]]][['y']][,data.names]
          
          prior <- model.vintages[[taus[t]]]$prior
          if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
          c <- which(prior$state.names=='c') # Index for cycle
          
          model <- list()
          model$F <- model.vintages[[i]]$build.F(prior, param)
          model$G <- model.vintages[[i]]$build.G(prior, param)
          model$H <- model.vintages[[i]]$build.H(prior)
          model$omega <- diag(param$omega)
          
          K <- KFKS(data$y, prior, model)
          w <- w.stt(data$y, prior, model, K, dim(K$KS$stT)[1])
          
          W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
          temp_draws[t, ,i_draws] <- cbind(W, w$stt)[c,]
        }
      }
    }
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')        
        
  } else {
    cat("\n ")
    cat("real time \n")
    cat("\n ")
    
    for (i_draws in 1:N_draws) {
      for (t in 1:length(taus)) {
        S = model.vintages[[taus[t]]]$res$S[,4,i_draws]
        param = list()
        param$phi = model.vintages[[taus[t]]]$res$phi[,i_draws]
        param$omega = model.vintages[[taus[t]]]$res$omega[,i_draws]
        if (length(model.vintages[[taus[t]]]$res$alpha[,i_draws]) != 0) param$alpha = model.vintages[[taus[t]]]$res$alpha[,i_draws]
        if (length(model.vintages[[taus[t]]]$res$beta[,i_draws]) != 0) param$beta = model.vintages[[taus[t]]]$res$beta[,i_draws]
          if (length(S)!=0) {
            
            data$y <- data.vintages[[taus[t]]][['y']][,data.names]
          
            prior <- model.vintages[[taus[t]]]$prior
            if (prior$T != dim(data$y)[1]) prior$T <- dim(data$y)[1] 
            c <- which(prior$state.names=='c') # Index for cycle
    
            model <- list()
            model$F <- model.vintages[[i]]$build.F(prior, param)
            model$G <- model.vintages[[i]]$build.G(prior, param)
            model$H <- model.vintages[[i]]$build.H(prior)
            model$omega <- diag(param$omega)
            
            K <- KFKS(data$y, prior, model)
            w <- w.stT(data$y, prior, model, K, dim(K$KS$stT)[1])
            
            W <- apply(w$cont, c(1,2), sum) + matrix(c(w$w0, rep(0, prior$Ns*(prior$Nobs-1))), ncol=prior$Nobs)
            temp_draws[t, ,i_draws] <- cbind(W, w$stT)[c,]
          }
        }
      }
    temp_contr <- apply(temp_draws, c(1,2), mean)
    # This is storage for the decomposition
    contr <- xts(array(temp_contr, c(length(taus), length(data.names)+1)), 
                 order.by=as.yearqtr(as.Date(vintage.dates[taus]) %m+% -months(3)))
    colnames(contr) <- c(data.names, 'c')
    
  }
  return(contr)
}

tau <- '1965 Q3' # Lower bound for the dates

# UC2 (Y/Pi) -------------
rt.uc2.pi <- readRDS('../Raw model output/rt-uc2.pi.Rds')
qrt.uc2.pi <- readRDS('../Raw model output/qrt-uc2.pi.Rds')

start_time_uc2pi <- Sys.time()
uc2.pi <- list()
uc2.pi$rt <- track.contribution(rt.vintages, rt.uc2.pi, final=FALSE, quasifinal=FALSE, quasireal=FALSE, tau)
uc2.pi$qrt <- track.contribution(qrt.vintages, qrt.uc2.pi, final=FALSE, quasifinal=FALSE, quasireal=TRUE, tau)
uc2.pi$qf <- track.contribution(qrt.vintages, qrt.uc2.pi,final=FALSE, quasifinal=TRUE, quasireal=FALSE, tau)
uc2.pi$final <- track.contribution(rt.vintages, rt.uc2.pi, final=TRUE, quasifinal=FALSE, quasireal=FALSE, tau)
end_time_uc2pi <- Sys.time()
end_time_uc2pi - start_time_uc2pi
print(end_time_uc2pi - start_time_uc2pi)

rm(rt.uc2.pi, qrt.uc2.pi)

write.table(uc2.pi, "../Decomposition/uc2pi_decomp.txt", sep="\t") 
