rm(list=ls())
require(xts)
require(lubridate)
require(tis)

rm(list=ls())
local.dir <- '../Data/'
data.loc <- local.dir
setwd(local.dir)

load('Vintage-datasets-Nov-2023.Rda') # Produced in wrangle-vintage-data.R
vintage.dates <- as.yearqtr(as.Date(sapply(rt.vintages, function(x) x[['vintage']])))
rm(rt.vintages, qrt.vintages)

# Tealbook ------
tb <- xts(order.by=as.yearqtr(seq(from=as.Date('1965-07-01'), # First RT gap 
                                  to=as.Date('2022-01-01'), # Last RT gap (all models)
                                  by='quarter')))
# Edge-Rudd data -- already lagged!
raw <- read.csv('gbgap_er.csv')
er <- xts(raw[,-1], order.by=as.yearqtr(seq(from=as.Date('1950-10-01'), length.out=dim(raw)[1], by='quarter')))
tb <- merge(tb, er['1965-07-01/'])
colnames(tb) <- 'er'

# This is from Philly fed
tb.frb.ph <- xts(read.csv('Greenbook_Output_Gap_Aug_2022.csv')[,-1], 
           order.by=as.yearqtr(seq(from=as.Date('1975-01-01'), 
                                   to=as.Date('2019-10-01'), 
                                   by='quarter')))
. <- substr(colnames(tb.frb.ph), 7,12)
tb.dates <- as.Date(., '%y%m%d')

# This loop picks off, for each Tealbook date, the t-1 output gap estimate
tb.gaps <- xts(array(NA, length(tb.dates)), order.by=tb.dates)
for (i in 1:(length(tb.dates))) {
  tm1 <- as.yearqtr(tb.dates[i] %m-% months(3))
  tb.gaps[i,1] <- tb.frb.ph[tm1,i] # NOTE TIMING MATCHES ER csv file
}

# And now select the Tealbook that is first in the quarter
raw <- cbind(tb.gaps, I=0)
n <- dim(raw)[1]
raw$qtr <- quarter(index(raw))
raw$I[which(as.vector(raw$qtr[2:n])==as.vector(raw$qtr[1:(n-1)]))] <- 1
rt <- xts(raw[which(raw$I==1), 1], 
          order.by=seq(from=as.Date('1996-01-01'), length.out=length(which(raw$I==1)), by='quarter'))
colnames(rt) <- NULL
merge(tb, rt) # Checking that the overlapping dates are the same

# Replace recent values in tb
tb['2008-10-01/2016-07-01'] <- rt['2008-10-01/2016-07-01']

# And fill in QRT, final
tb <- merge(tb, qrt=NA) # There are none!
final.2016 <- tb.frb.ph[,dim(tb.frb.ph)[2]]
tb <- merge(tb, final.2016['1965-10-01/2016'])
tb <- merge(tb, final=NA) # There are none!

colnames(tb) <- c('rt', 'qrt', 'final.2016', 'final')
rm(list=setdiff(ls(), c('vintage.dates', 'tb', 'local.dir', 'data.loc')))


# Univariate gaps ----------------
gather.uni.data <- function(rt, qrt, vintage.dates) {

  V <- length(vintage.dates)
  X <- xts(array(NA, c(V, 4)), 
           order.by=as.yearqtr(as.Date(vintage.dates) %m+% -months(3)))
  colnames(X) <- c('rt', 'qrt', 'final.2016', 'final')
  
  # Fill in real-time vintages
  for (v in 1:V) {
    tm1 <- as.yearqtr(as.Date(colnames(rt)[v]) %m-% months(3))
    X[tm1, 'rt'] <- rt[tm1,v]
  }
  
  # Fill in quasi real-time vintages
  for (v in 1:V) {
    tm1 <- as.yearqtr(as.Date(colnames(qrt)[v]) %m-% months(3))
    X[tm1, 'qrt'] <- qrt[tm1,v]
  }
  
  # Fill in final vintages
  V.2016 <- which(as.yearqtr(as.Date(colnames(rt))) == "2016 Q4")
  start <- as.Date(vintage.dates[1]) %m+% -months(3)
  X[,'final.2016'] <- rt[paste0(start,'/'), V.2016]
  X[,'final'] <- rt[paste0(start,'/'), V]

  return(X)
} # Function to produce useful matrix

uni.raw.rt <- readRDS('uni-rt-gap-vintages.Rds')
uni.raw.qrt <- readRDS('uni-qrt-gap-vintages.Rds')

uni <- mapply(function(x,y) gather.uni.data(x, y, vintage.dates), x=uni.raw.rt, y=uni.raw.qrt, SIMPLIFY = FALSE)


# UC gaps -----
from.raw.output <- FALSE
if (from.raw.output) { # The files are huge so reading the raw stuff takes some time...saved separately within this chunk then called if "else" 
  rm(list=setdiff(ls(), c('tb', 'uni', 
                          'local.dir', 'data.loc', 
                          'vintage.dates')))
  gather.uc.output <- function(rt, qrt, vintage.dates) {
  
    V <- length(vintage.dates)
    rt.qtrs <- as.Date(vintage.dates) %m+% -months(3)
    c <- which(qrt[[1]][['prior']][['state.names']]=='c')  # An index for the cycle
  
    X <- xts(array(NA, c(V, 4)), 
             order.by=as.yearqtr(rt.qtrs))
    colnames(X) <- c('rt', 'qrt', 'final.2016', 'final')
    
    
    for (v in 1:V) {
      tm1 <- as.yearqtr(as.Date(vintage.dates[v]) %m+% -months(3))
      
      if ((v==1 && length(rt[[v]])==6) | length(rt[[v]])==2) {
        # Fill in real-time vintages
        S.rt <- apply(rt[[v]][['res']][['S']], c(1,2), mean)
        X[tm1, 'rt'] <- last(S.rt)[c]
      }
        
      # Fill in quasi real-time vintages
      S.qrt <- apply(qrt[[v]][['res']][['S']], c(1,2), mean)
      X[tm1, 'qrt'] <- last(S.qrt)[c]
    }
  
    # Fill in 2016Q4 vintage
    V.2016 <- which(vintage.dates == '2016 Q4')
    . <- apply(rt[[V.2016]][['res']][['S']], c(1,2), mean)
    S.v2016 <- xts(., order.by=as.yearqtr(seq(from=as.Date('1960-01-01'), 
                                   length.out=dim(.)[1], 
                                   by='quarter')))['1965-07-01/',]
    X[,'final.2016'] <- c(matrix(S.v2016[,c]), 
                          rep(NA, (V-dim(S.v2016)[1])))
    
    # Fill in 2019Q1 vintage
    . <- apply(rt[[V]][['res']][['S']], c(1,2), mean)
    S.V <- xts(., order.by=as.yearqtr(seq(from=as.Date('1960-01-01'), 
                                              length.out=dim(.)[1], 
                                              by='quarter')))['1965-07-01/',]  
    X[,'final'] <- matrix(S.V[,c])
    return(X)
  }
  
  uc.gaps <- list()
  
  # UC1 
  uc1.rt <- readRDS(paste0(data.loc, 'rt-uc1.Rds'))
  uc1.qrt <- readRDS(paste0(data.loc, 'qrt-uc1.Rds'))
  uc.gaps$uc1 <- gather.uc.output(uc1.rt, uc1.qrt, vintage.dates)
  rm(uc1.rt, uc1.qrt)

  # UC1u 
  uc1u.rt <- readRDS(paste0(data.loc, 'rt-uc1u.Rds'))
  uc1u.qrt <- readRDS(paste0(data.loc, 'qrt-uc1u.Rds'))
  uc.gaps$uc1u <- gather.uc.output(uc1u.rt, uc1u.qrt, vintage.dates)
  rm(uc1u.rt, uc1u.qrt)
  
  # UC2 (y/pi) 
  uc2.rt <- readRDS(paste0(data.loc, 'rt-uc2.pi.Rds'))
  uc2.qrt <- readRDS(paste0(data.loc, 'qrt-uc2.pi.Rds'))
  uc.gaps$uc2.pi <- gather.uc.output(uc2.rt, uc2.qrt, vintage.dates)
  rm(uc2.rt, uc2.qrt)
  
  
  # UC2 (y/cu) 
  uc2.rt <- readRDS(paste0(data.loc, 'rt-uc2.cu.Rds'))
  uc2.qrt <- readRDS(paste0(data.loc, 'qrt-uc2.cu.Rds'))
  uc.gaps$uc2.cu <- gather.uc.output(uc2.rt, uc2.qrt, vintage.dates)
  rm(uc2.rt, uc2.qrt)
  
  
  # UC2 (y/u) 
  uc2.rt <- readRDS(paste0(data.loc, 'rt-uc2.u.Rds'))
  uc2.qrt <- readRDS(paste0(data.loc, 'qrt-uc2.u.Rds'))
  uc.gaps$uc2.u <- gather.uc.output(uc2.rt, uc2.qrt, vintage.dates)
  rm(uc2.rt, uc2.qrt)
  
  
  # UC3 
  uc3.rt <- readRDS(paste0(data.loc, 'rt-uc3.Rds'))
  uc3.qrt <- readRDS(paste0(data.loc, 'qrt-uc3.Rds'))
  uc.gaps$uc3 <- gather.uc.output(uc3.rt, uc3.qrt, vintage.dates)
  rm(uc3.rt, uc3.qrt)
  
  saveRDS(uc.gaps, 'UC_vintage_gaps.Rds')

  } else {
    uc.gaps <- readRDS('UC_vintage_gaps.Rds')
  }


# MVBN --------------------

raw <-read.csv('MW_multivariateBN_constant_lambda.csv', sep=',', blank.lines.skip=FALSE)
raw_final = read.csv('MW_multivariateBN_constant_lambda_final.csv')
raw_2016Q4 = read.csv('MW_multivariateBN_constant_lambda_2016Q4.csv')

V <- length(vintage.dates)
mv.bn <- xts(array(NA, c(V, 4)), 
             order.by=as.yearqtr(as.Date(vintage.dates) %m+% -months(3)))
colnames(mv.bn) <- c('rt', 'qrt', 'final.2016', 'final')

mv.bn[,"rt"] <- array(raw)[[1]]
mv.bn[,"final"] <- array(raw_final)[[1]]
mv.bn[,"final.2016"] <- array(raw_2016Q4)[[1]]

raw_pi <-read.csv('MW_multivariateBN_constant_lambda_pi.csv', sep=',', blank.lines.skip=FALSE)
raw_final_pi = read.csv('MW_multivariateBN_constant_lambda_final_pi.csv')
raw_2016Q4_pi = read.csv('MW_multivariateBN_constant_lambda_2016Q4_pi.csv')

mv.bn_tri <- xts(array(NA, c(V, 4)), 
                 order.by=as.yearqtr(as.Date(vintage.dates) %m+% -months(3)))
colnames(mv.bn_tri) <- c('rt', 'qrt', 'final.2016', 'final')

mv.bn_tri[,"rt"] <- array(raw_pi)[[1]]
mv.bn_tri[,"final"] <- array(raw_final_pi)[[1]]
mv.bn_tri[,"final.2016"] <- array(raw_2016Q4_pi)[[1]]


# CBO ---------
substr.right <- function(x,n) {
  substr(x, nchar(x)-n+1, nchar(x))
}

# Vintages of nominal potential GDP
. <- data.frame(readxl::read_excel('All-CBO-nominal-potential.xlsx'))
npot <- xts(.[,-1], seq(from=as.Date(as.yearqtr('1949 Q1')),
                        to=as.Date(as.yearqtr('2032 Q4')), 
                        by='quarter'))
colnames(npot) <- paste(substr(colnames(npot), start=1, stop=3), substr.right(colnames(npot), 4))

# Vintages of nominal GDP
. <- data.frame(readxl::read_excel('noutputMvQd.xlsx'))
ngdp <- xts(.[,-1], seq(from=as.Date(as.yearqtr('1947 Q1')),
                        to=as.Date(as.yearqtr('2022 Q2')), 
                        by='quarter'))
colnames(ngdp) <- paste0(as.yearmon(seq(from=as.Date('1965-11-01'), to=as.Date('2022-08-01'), by='month')))

# NGDP forecasts from CBO when available
. <- data.frame(readxl::read_excel('All-CBO-GDP-forecast.xlsx'))
ngdp.hat <- xts(.[,-1], seq(from=as.Date(as.yearqtr('2009 Q1')),
                            to=as.Date(as.yearqtr('2032 Q4')), 
                            by='quarter'))
colnames(ngdp.hat) <- paste(substr(colnames(ngdp.hat), start=1, stop=3), substr.right(colnames(ngdp.hat), 4))


# Create quarterly vintages since Feb 1991
index <- as.yearqtr(seq(as.Date(as.yearqtr('1947 Q1')),
                        as.Date(as.yearqtr('2032 Q4')), 
                        by='quarter'))
vintages <- as.yearmon(seq(from=as.Date('1991-02-01'), to=as.Date('2022-05-01'), by='quarter'))
cbo.gaps <- xts(array(NA, c(length(index), length(vintages))), order.by=index)
colnames(cbo.gaps) <- paste(as.yearmon(vintages))

for (v in 1:length(vintages)) {
  
  # Column selectors for pot, ngdp
  c.pot <- max(which(as.yearmon(colnames(npot))<=vintages[v]))
  c.ngdp <- which(colnames(ngdp) == vintages[v])
  
  if (vintages[v] < as.yearmon(colnames(ngdp.hat)[1])) {
    
    X <- cbind(npot[,c.pot], ngdp[,c.ngdp]); colnames(X) <- c('pot','ngdp') # cbind just to be sure dates line up
    cbo.gaps[,v] <- 100*(log(X[,'ngdp'] / X[,'pot']))
    
  } else {
    
    c.ngdp.hat <- max(which(colnames(ngdp.hat) <= as.yearmon(vintages[v])))  # Column of ngdp.hat
    X <- cbind(npot[,c.pot], ngdp[,c.ngdp], ngdp.hat[,c.ngdp.hat], ngdp.hat[,c.ngdp.hat] / stats::lag(ngdp.hat[,c.ngdp.hat])) 
    colnames(X) <- c('pot','ngdp','ngdp.hat', 'gr')
    
    # Grow ngdp forward
    last <- max(which(!is.na(X[,'ngdp'])))
    for (i in (last+1):dim(X)[1]) {
      X[i,'ngdp'] <- matrix(X[(i-1),'ngdp'])*matrix(X[i,'gr'])
    }
    cbo.gaps[,v] <- 100*(log(X[,'ngdp'] / X[,'pot']))
  }
}



# Now gather into the "usual" format
cbo <- xts(array(NA, c(length(vintage.dates), 4)), 
           order.by=as.Date(vintage.dates) %m-% months(3))
colnames(cbo) <- c('rt','qrt','final.2016','final')

for (v in 1:length(vintage.dates)) {
  if (vintage.dates[v] %in% as.yearqtr(as.yearmon(colnames(cbo.gaps)))) {
    tm1 <- as.yearqtr(as.Date(vintage.dates[v]) %m-% months(3)) # Quarter
    i <- which(as.yearqtr(as.yearmon(colnames(cbo.gaps))) == as.yearqtr(vintage.dates[v])) # Column
    cbo[tm1, 'rt'] <- cbo.gaps[tm1,i]
  }
}

# QRT do not exist!
# Final vintages
V.2016 <- which(as.yearqtr(as.yearmon(colnames(cbo.gaps))) == "2016 Q4")
V.2022 <- which(as.yearqtr(as.yearmon(colnames(cbo.gaps))) == "2022 Q2")
start <- as.Date(vintage.dates[1]) %m+% -months(3)
end <- as.Date(vintage.dates[length(vintage.dates)]) %m+% -months(3)
cbo[,'final.2016'] <- cbo.gaps[paste0(start,'/',end), V.2016]
cbo[,'final'] <- cbo.gaps[paste0(start,'/',end), V.2022]


# Put them together --------------
gaps <- c(uni, uc.gaps)
gaps[['tb']] <- tb
gaps[['cbo']] <- cbo
gaps[['mv.bn']] <- mv.bn
gaps[['mv.bn_tri']] <- mv.bn_tri

gaps <- gaps[c("line", "quad", "hp", "hp.exp", "ham", "mw", "kmw", "mv.bn", "mv.bn_tri",
               "uc1","uc1u", "uc2.pi", "uc2.cu", "uc2.u", "uc3", "tb", 'cbo')]

# Fill in RT as missing for a few of the models
t <- which(index(gaps[[1]])=='1995 Q4')
gaps[['hp.exp']][t,'rt'] <- NA
gaps[['uc1u']][t,'rt'] <- NA
gaps[['uc2.u']][t,'rt'] <- NA
gaps[['uc2.cu']][t,'rt'] <- NA
gaps[['uc3']][t,'rt'] <- NA

sapply(gaps, function(x) sum(is.na(x)))


# Save -----
save <- TRUE
if (save) {
  saveRDS(gaps, file=paste0(data.loc, 'All-gap-vintages-Nov-2023.Rds'))
}